unit XKeyboardEvents;

interface

uses Windows, Messages, XClasses, XVisuals, XForms, XControls;

type
   XOnFormKey = procedure( Sender : TObject; var Key : Word; Shift : XShiftState )
                of object;
                
   XFormsStdKeyEvents = class
   private
      fOldOnMessage : XOnFormMessage;
      fOnKeyDown : XOnFormKey;
      procedure HandleKey( Sender : TObject; var Msg : TMsg; var Rslt : Integer; var Handled : Boolean );
   protected
      procedure FindTabControls( F : XVisual; C : XControl; var First, Prev, Next, Last : XControl );
                dynamic;
      function GotoControl( F : XForm; Key : DWord ) : Boolean;
      procedure DoGotoControl( Sender : TObject; Key : DWord );
   public
      constructor Create;
      property On_KeyDown : XOnFormKey read fOnKeyDown write fOnKeyDown;
   end;

var FormsStdKeyEvents : XFormsStdKeyEvents;

procedure UseKeyboard;
// Call UseKeyboard to provide keyboard handling messages adding
// about 1,7Kbytes code to executable.

implementation

procedure UseKeyboard;
begin
   if FormsStdKeyEvents <> nil then Exit;
   FormsStdKeyEvents := XFormsStdKeyEvents.Create;
end;

{ XFormsStdKeyEvents }

constructor XFormsStdKeyEvents.Create;
begin
   inherited;
   if GlobalFormsManager = nil then
      GlobalFormsManager := XFormsManager.Create;
   fOldOnMessage := GlobalFormsManager.On_FormMessage;
   GlobalFormsManager.On_FormMessage := HandleKey;
   GlobalFormsManager.OnGotoControl := DoGotoControl;
end;

procedure XFormsStdKeyEvents.DoGotoControl(Sender: TObject; Key: DWord);
begin
   GotoControl( Sender as XForm, Key );
end;

procedure XFormsStdKeyEvents.FindTabControls( F : XVisual; C : XControl; var First, Prev, Next, Last : XControl );
var I : Integer;
    E : XClass;
    X : XControl;
begin
   for I := 0 to F.ChildsCount - 1 do
   begin
      E := F.Childs[ I ];
      if E is XControl then
      begin
         X := E as XControl;
         if not X.Showing then continue;
         if not X.CanFocus then continue;
         FindTabControls( E as XVisual, C, First, Prev, Next, Last );
         if X.TabOrder = 0 then continue;
         if (First = nil) or (First.TabOrder > X.TabOrder) then
            First := X;
         if (Last = nil) or (Last.TabOrder < X.TabOrder) then
            Last := X;
         if C <> X then
         begin
            if ((C = nil) or (X.TabOrder < C.TabOrder))
               and ((Prev = nil) or (X.TabOrder > Prev.TabOrder)) then
               Prev := X;
            if ((C = nil) or (X.TabOrder > C.TabOrder))
               and ((Next = nil) or (X.TabOrder < Next.TabOrder)) then
               Next := X;
         end;
      end;
   end;
end;

function XFormsStdKeyEvents.GotoControl(F : XForm; Key: DWord)
         : Boolean;
var First, Prev, Next, Last : XControl;
begin
   First := nil; Prev := nil; Next := nil; Last := nil;
   FindTabControls( F, F.CurrentControl, First, Prev, Next, Last );
   Result := True;
   case Key of
   VK_TAB, VK_NEXT, VK_RIGHT, VK_DOWN:
      if Next = nil then
         F.CurrentControl := First
      else
         F.CurrentControl := Next;
   VK_TAB or $80000000, VK_PRIOR, VK_LEFT, VK_UP:
      if Prev = nil then
         F.CurrentControl := Last
      else
         F.CurrentControl := Prev;
   VK_HOME:
      F.CurrentControl := First;
   VK_END:
      F.CurrentControl := Last;
   else
      Result := False;
   end;
end;

procedure XFormsStdKeyEvents.HandleKey(Sender: TObject; var Msg: TMsg;
  var Rslt: Integer; var Handled: Boolean);
var F : XForm;
    Shift : XShiftState;
    Key : DWord;
    WKey : Word;
    procedure CallOldMessage;
    begin
       if Assigned( fOldOnMessage ) then
          fOldOnMessage( Sender, Msg, Rslt, Handled );
    end;
    procedure GetKeyData;
    begin
       Shift := [];
       if GetKeyState( VK_SHIFT ) < 0 then Shift := [ sbShift ];
       if GetKeyState( VK_CONTROL ) < 0 then Include( Shift, sbCtrl );
       if (Msg.lParam and $20000000) <> 0 then Include( Shift, sbAlt );
       Key := Msg.wParam and $FFFF;
       WKey := Word( Key );
    end;
    function KeyDown : Boolean;
    begin
       Result := False;
       GetKeyData;
       if F.CurrentControl <> nil then
       begin
          F.CurrentControl.DoKeyDown( WKey, Shift );
          if WKey = 0 then
             Result := True;
       end;
       if ((Key = VK_TAB) and ((Shift = []) or (Shift = [sbShift])))
          or (Key in [VK_HOME,VK_END,VK_NEXT,VK_PRIOR,VK_RIGHT,VK_DOWN,VK_LEFT,VK_UP]) then
       begin
          if Shift <> [] then
             Key := Key or $80000000;
          Result := GotoControl( F, Key );
          Exit;
       end;
       if Assigned( fOnKeyDown ) then
       begin
          fOnKeyDown( F, WKey, Shift );
          Result := WKey = 0;
       end;
    end;
    function KeyUp : Boolean;
    begin
       Result := False;
       if F.CurrentControl = nil then Exit;
       GetKeyData;
       F.CurrentControl.DoKeyUp( WKey, Shift );
       Result := WKey = 0;
    end;
    function KeyChar : Boolean;
    var C : Char;
    begin
       Result := False;
       if F.CurrentControl = nil then Exit;
       GetKeyData;
       C := Char( Key );
       F.CurrentControl.DoKeyPress( C );
       Result := C = #0;
    end;
begin
   F := Sender as XForm;
   case Msg.message of
   WM_KEYDOWN, WM_SYSKEYDOWN :
      if not KeyDown then
      begin
         CallOldMessage;
         Exit;
      end;
   WM_KEYUP, WM_SYSKEYUP : if not KeyUp then
                           begin
                              CallOldMessage;
                              Exit;
                           end;
   WM_CHAR, WM_SYSCHAR: if not KeyChar then
                        begin
                           CallOldMessage;
                           Exit;
                        end;
   else begin
           CallOldMessage;
           Exit;
        end;
   end;
   Rslt := 0;
   Handled := True;
end;

end.
