unit XMfcAntiFlick;

interface

uses windows, messages, XClasses, XVisuals, XWindows, XForms, XMfcControls;

type
   XMfcAntiFlicker = class
   private
     FOldOnMessage : XOnFormMessage;
     FTextSetting : Boolean;
   protected
     procedure Execute( Sender : TObject; var Msg : TMsg; var Rslt : Integer; var Handled : Boolean );
   public
     constructor Create;
   end;

var MfcControlsAntiFlicker : XMfcAntiFlicker;

procedure UseMfcAntiFlicker;

implementation

type
  XMfcCover = class( XWCustomPaint )
  private
     FBounds : TRect;
     procedure DoNothing( Sender : TObject );
  protected
    procedure SetBoundsRect( Value : TRect ); override;
  public
    property OnPaint;
    property OnEraseBkgnd;
    constructor Create( AParent : XClass ); override;
    function CreateWindow : Boolean; override;
  end;

{ XMfcAntiFlicker }

constructor XMfcAntiFlicker.Create;
begin
  inherited;
  FOldOnMessage := GlobalMfcControlsManager.On_MfcControlMessage;
  GlobalMfcControlsManager.On_MfcControlMessage := Execute;
end;

procedure XMfcAntiFlicker.Execute(Sender: TObject; var Msg: TMsg;
  var Rslt: Integer; var Handled: Boolean);
var
  DC, MemDC: HDC;
  MemBitmap, OldBitmap: HBITMAP;
  PS: TPaintStruct;
  MC : XMfcControl;
  CR : TRect;
  Cover : XMfcCover;

  procedure DoPaintThroughBuffer;
  begin
     DC := BeginPaint( Msg.hwnd, PS );
     SendMessage( Msg.hwnd, WM_ERASEBKGND, MemDC, MemDC );
     SendMessage( Msg.hwnd, WM_Paint, MemDC, 0 );
     BitBlt( DC, 0, 0, CR.Right, CR.Bottom, MemDC, 0, 0, SRCCOPY );
     EndPaint( Msg.hwnd, PS);
  end;
begin
  case Msg.message of
  WM_PAINT:
    begin
      if Msg.wParam = 0 then
      begin
         MC := Sender as XMfcControl;
         // Painting through buffer to prevent flicks:
         DC := GetDC( 0 );
         CR := MC.ClientRect;
         MemBitmap := CreateCompatibleBitmap(DC, CR.Right, CR.Bottom);
         ReleaseDC(0, DC);
         MemDC := CreateCompatibleDC(0);
         OldBitmap := SelectObject(MemDC, MemBitmap);
         //try
         DoPaintThroughBuffer;
         //finally
         SelectObject(MemDC, OldBitmap);
         DeleteDC(MemDC);
         DeleteObject(MemBitmap);
         //end;
         Rslt := 0;
         Handled := True;
      end;
    end;
  WM_ERASEBKGND:
    if Msg.wParam = 0 then
    begin
      Rslt := 1;
      Handled := True;
    end;
  WM_SETTEXT:
    if not FTextSetting then
    begin
      // Here we have to cover control with XWPaint, which
      // ignores any painting, and then pass message through
      // and after that destroy covering object. This will
      // initiate normal paint without flickering.
      MC := Sender as XMfcControl;
      Cover := XMfcCover.Create( MC );
      Cover.BoundsRect := MC.ClientRect;
      Cover.CreateWindow;
      FTextSetting := True;
      Rslt := SendMessage( Msg.hwnd, Msg.message, Msg.wParam, Msg.lParam );
      FTextSetting := False;
      Cover.Free;
      MC.Invalidate;
      Handled := True;
    end;
  else
    if Assigned( FOldOnMessage ) then
       FOldOnMessage( Sender, Msg, Rslt, Handled );
  end;
end;

procedure UseMfcAntiFlicker;
begin
   if MfcControlsAntiFlicker = nil then
      MfcControlsAntiFlicker := XMfcAntiFlicker.Create;
end;

{ XMfcCover }

constructor XMfcCover.Create(AParent: XClass);
begin
  inherited;
  OnEraseBkgnd := DoNothing;
end;

function XMfcCover.CreateWindow: Boolean;
begin
   Result := inherited CreateWindow;
   SetBoundsRect( FBounds );
end;

procedure XMfcCover.DoNothing(Sender: TObject);
begin
  // nothing to do instead of erase background
end;

procedure XMfcCover.SetBoundsRect(Value: TRect);
var Handle : HWnd;
begin
   Handle := Self.GetWindowHandle;
   FBounds := Value;
   if Handle <> 0 then
   SetWindowPos( Handle, 0, Value.Left, Value.Top, Value.Right - Value.Left,
                 Value.Bottom - Value.Top, SWP_NOZORDER );
end;

end.
