unit XMouseEvents;

interface

uses Windows, Messages, XClasses, XForms, XControls;

type

   XFormsStdMouseEvents = class( TObject )
   private
      fOldOnMessage : XOnFormMessage;
      procedure HandleMessage( Sender : TObject; var Msg : TMsg;
                               var Rslt : Integer; var Handled : Boolean );
   protected
   public
      constructor Create;
   end;

var FormsStdMouseEvents : XFormsStdMouseEvents;

procedure UseStdMouse;
// Call UseStdMouse anywhere before running applet (the best
// place is in dpr-file before creating main form). Refusing
// from standard mouse events handling in forms derived from
// XForm saves about 1,1Kb of code in resulting executable. 

implementation

procedure UseStdMouse;
begin
   if FormsStdMouseEvents <> nil then Exit;
   FormsStdMouseEvents := XFormsStdMouseEvents.Create;
end;

constructor XFormsStdMouseEvents.Create;
begin
   inherited;
   if GlobalFormsManager = nil then
      GlobalFormsManager := XFormsManager.Create;
   fOldOnMessage := GlobalFormsManager.On_FormMessage;
   GlobalFormsManager.On_FormMessage := HandleMessage;
end;

procedure XFormsStdMouseEvents.HandleMessage( Sender : TObject; var Msg : TMsg;
                                              var Rslt : Integer; var Handled : Boolean );
var F : XForm;
    C : XControl;
    Shift : XShiftState;
    X, Y : Integer;
   procedure GetMouseEventParams;
   begin
      Shift := [];
      if (Msg.wParam and MK_CONTROL) <> 0 then
         Shift := Shift + [sbCtrl];
      if (Msg.wParam and MK_LBUTTON) <> 0 then
         Shift := Shift + [sbLeft];
      if (Msg.wParam and MK_MBUTTON) <> 0 then
         Shift := Shift + [sbMiddle];
      if (Msg.wParam and MK_RBUTTON) <> 0 then
         Shift := Shift + [sbRight];
      if (Msg.wParam and MK_SHIFT) <> 0 then
         Shift := Shift + [sbShift];
      if GetKeyState( VK_MENU ) < 0 then
         Shift := Shift + [sbAlt];
      X := LOWORD( Msg.lParam );
      Y := HIWORD( Msg.lParam );
      C := F.ControlAtPos( X, Y, True );
   end;
   procedure MouseButtonDown( Button : XMouseButton );
   begin
      GetMouseEventParams;
      if C <> nil then
         C.DoMouseDown( Button, Shift, X - C.Left, Y - C.Top )
      else
      if Assigned( F.OnMouseDown ) then
         F.OnMouseDown( Self, Button, Shift, X, Y );
   end;
   procedure MouseButtonUp( Button : XMouseButton );
   begin
      GetMouseEventParams;
      if GetCapture = F.Handle then
      begin
         if F.CurrentControl <> nil then
         begin
            F.CurrentControl.DoMouseUp( Button, Shift,
                             X - F.CurrentControl.ControlRect.Left,
                             Y - F.CurrentControl.ControlRect.Top );
            C := nil;
         end;
         ReleaseCapture;
      end;
      if C <> nil then
         C.DoMouseUp( Button, Shift, X - C.Left, Y - C.Top )
      else
      if Assigned( F.OnMouseUp ) then
         F.OnMouseUp( Self, Button, Shift, X, Y );
   end;
begin
   F := Sender as XForm;
   case Msg.message of
   WM_MOUSEMOVE: begin
                    GetMouseEventParams;
                    if C <> nil then
                       C.DoMouseMove( Shift, X - C.Left, Y - C.Top )
                    else
                    if Assigned( F.OnMouseMove ) then
                       F.OnMouseMove( F, Shift, X, Y );
                 end;
   WM_LBUTTONDOWN: MouseButtonDown( mbLeft );
   WM_MBUTTONDOWN: MouseButtonDown( mbMiddle );
   WM_RBUTTONDOWN: MouseButtonDown( mbRight );
   WM_LBUTTONUP  : MouseButtonUp( mbLeft );
   WM_MBUTTONUP  : MouseButtonUp( mbMiddle );
   WM_RBUTTONUP  : MouseButtonUp( mbRight );
   else begin
          if Assigned( fOldOnMessage ) then
             fOldOnMessage( Sender, Msg, Rslt, Handled );
          Exit;
        end;
   end;
   Rslt := 0;
   Handled := True;
end;

end.
